from .parser import parse
from .exceptions import SNSinteractionException, SNSserverException
from .records import SNSuserRecord
from .objects import SNSinteraction, SNScontent

import socket
import re

SNS_PORT = 3735

#TODO: encrypt the whole request with the private key of the user first, to ensure integrity and authenticity. 
#Since the public key is already known to the server, this does not require any extra steps either.
def request(type, action, content=None, **headers):
    for key, value in headers.items():
        try:
            headers[key] = str(value)
        except IndexError:
            raise SNSinteractionException(str(headers), "One of the interaction headers could not be parsed to a string")

    interaction = parse(str(SNSinteraction(type, action, headers, content)))

    if interaction.headers['To'].count('@') > 1:
        raise SNSinteractionException(str(interaction), "The To header contains too many @ characters (max. 1 per address header), causing that the address cannot be determined")

    try:
        address = interaction.headers['To'].split('@')[1]
    except:
        address = SNSuserRecord.provider

    conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    conn.connect((address, SNS_PORT))
    conn.sendall(str(interaction).encode('utf-8'))

    buffer_size = 4096
    data = b''

    while True:
        temp_data = conn.recv(buffer_size)
        data += temp_data
        if (len(temp_data) < buffer_size): break
        time.sleep(0.2)

    conn.close()

    interaction_blobs = re.split(r"(?:\r?\n){2,}", data.decode('utf-8'))

    interactions = []
    for interaction_blob in interaction_blobs:
        try:
            interactions.append(parse(interaction_blob))
        except:
            if len(interactions) > 0:
                interactions[-1].content.append(interaction_blob)
            else:
                raise SNSserverException()
    
    return interactions

def user(action, To, content=None, **headers):
    encrypted_content = content

    headers['From'] = 'user:' + SNSuserRecord.identifier
    headers['To'] = To

    interaction_object = request('USER', action, encrypted_content, **headers)
    
    return interaction_object

def group(action, To, content=None, **headers):
    encrypted_content = content

    headers['From'] = 'user:' + SNSuserRecord.identifier
    headers['To'] = To

    interaction_object = request('GROUP', action, encrypted_content, **headers)
    
    return interaction_object

def post(action, To, content=None, **headers):
    encrypted_content = content

    headers['From'] = 'user:' + SNSuserRecord.identifier
    headers['To'] = To

    interaction_object = request('POST', action, encrypted_content, **headers)
    
    return interaction_object

def reaction(action, To, content=None, **headers):
    encrypted_content = content

    headers['From'] = 'user:' + SNSuserRecord.identifier
    headers['To'] = To

    interaction_object = request('REACTION', action, encrypted_content, **headers)
    
    return interaction_object


import socket
import threading
import time
import re

from .parser import parse
from .objects import SNSinteraction
from .exceptions import SNSserverException

class SNSserver(object):

    HOST = '0.0.0.0'
    PORT = 3735

    components = {}

    server_thread = None
    conn_threads = []

    def __init__(self, type=None, action=None):
        self.type = type
        self.action = action

        if self.type is None:
            self.type = "*"

        if self.action is None:
            self.action = "*"

    def __call__(self, func):
        if not isinstance(self.type, list):
            self.type = [self.type]

        if not isinstance(self.action, list):
            self.action = [self.action]

        for type in self.type:
            for action in self.action:
                self.components[type, action] = func

    @staticmethod
    def start_server():
        SNSserver.server_thread = threading.Thread(target=SNSserver.server_thread)
        SNSserver.server_thread.start()

    @staticmethod
    def server_thread():
        print("Starting on:", SNSserver.HOST, SNSserver.PORT)
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind((SNSserver.HOST, SNSserver.PORT))
        s.listen(1)

        print("Server has started..")

        while True:
            conn, addr = s.accept()

            conn_thread = threading.Thread(target=SNSserver.connection_handler, args=(conn, addr))
            conn_thread.start()
            
            SNSserver.conn_threads.append(conn_thread)

        s.close()

    @staticmethod
    def connection_handler(conn, addr):
        with conn:
            print('Connected with', addr)
            conn.settimeout(2)
            while True:
                try:
                    buffer_size = 4096
                    data = b''

                    while True:
                        temp_data = conn.recv(buffer_size)
                        data += temp_data
                        if (len(temp_data) < buffer_size): break
                        time.sleep(0.2)

                    if len(data) == 0:
                        break

                    interaction_blobs = re.split(r"(?:\r?\n){2,}", data.decode('utf-8'))

                    interactions = []
                    for interaction_blob in interaction_blobs:
                        try:
                            interactions.append(parse(interaction_blob))
                        except:
                            if len(interactions) > 0:
                                interactions[-1].content.append(interaction_blob)
                            else:
                                raise SNSserverException()

                    print(addr, "Received interaction:")
                    print(interactions)

                    responses = []

                    for interaction_object in interactions:
                        response = None
                        if (interaction_object.type, interaction_object.action) in SNSserver.components:
                            response = SNSserver.components[interaction_object.type, interaction_object.action](interaction_object)
                        elif (interaction_object.type, "*") in SNSserver.components:
                            response = SNSserver.components[interaction_object.type, "*"](interaction_object)
                        elif ("*", interaction_object.action) in SNSserver.components:
                            response = SNSserver.components["*", interaction_object.action](interaction_object)
                        elif ("*", "*") in SNSserver.components:
                            response = SNSserver.components["*", "*"](interaction_object)
                        else:
                            print(addr, "Interaction component not found")
                            response = SNSinteraction("STATUS", 404, {'From':interaction_object.headers['To'], 'To':interaction_object.headers['From']})

                        if response is None:
                            continue
                        elif isinstance(response, list):
                            responses += response
                        else:
                            responses.append(response)

                    if len(responses) == 0:
                        raise SNSserverException()

                    print(addr, "Constructed response:")
                    if len(responses) == 0:
                        response = SNSinteraction("STATUS", 401, {'From':responses[0].headers['To'], 'To':responses[1].headers['From']})
                        print(response.type, response.action)
                        print(response.headers)
                        conn.sendall(str(response).encode('utf-8'))
                    elif len(responses) == 1:
                        response = responses[0]
                        print(response.type, response.action)
                        print(response.headers)
                        conn.sendall(str(response).encode('utf-8'))
                    else:
                        print('\n'.join([item.type + ' ' + item.action + '\n' + item.headers for item in responses]))
                        conn.sendall('\r\n\r\n'.join([str(item) for item in responses]).encode('utf-8'))

                except socket.error:
                    print("[WARNING] Something was wrong with the socket connection, further investigation is needed if this issue persists")
                    break
                except SNSserverException as e:
                    print(e)
                    print("[WARNING] A server exception occured, meaning something went wrong with loading in or sending a request")
                    break
        conn.close()


from .exceptions import SNSobjectException

class SNSinteraction:

    def __init__(self, type, action, headers={}, content=None):
        self.type = str(type)
        self.action = str(action)
        self.headers = headers
        self.content = SNScontent(content)

    def get(self, key, default=None):
        return self.headers.get(key, default)

    def __str__(self):
        interaction_string = self.type + " " + self.action

        if len(self.headers) > 0:
            interaction_string += '\r\n'
            interaction_string += '\r\n'.join([header[0] + ": " + str(header[1]) for header in self.headers.items()])

        content_string = str(self.content)

        if len(content_string) > 0:
            return interaction_string + '\r\n\r\n' + content_string
        else:
            return interaction_string

class SNScontent:

    def __init__(self, content=None):
        self.content = []

        self.append(content)

    def append_list(self, content):
        for item in content:
            self.append(item)

    def append(self, content):
        if type(content) == str:
            # try to parse, since it's probably decrypted content
            try:
                content_parameters = content.splitlines()
                _interaction_content = {}

                for content_parameter in content_parameters:
                    param_type, param_value = content_parameter.split(": ")

                    if param_type in _interaction_content:
                        _interaction_content[param_type] = [_interaction_content[param_type], param_value]
                    else:
                        _interaction_content[param_type] = param_value

                self.content.append(_interaction_content)
            except:
                raise SNSobjectException(content, "Content was passed as a str, but could not be parsed to a dict")
        elif isinstance(content, dict):
            # decrypted content
            self.content.append(content)
        elif type(content) == bytes:
            # could be encrypted content, by parsing we can find out
            self.append(content.decode('utf-8'))
        elif type(content) == SNScontent:
            # somehow we got an SNScontent object, just unpack and add to this class
            self.append_list(content.content)
        elif isinstance(content, list):
            self.append_list(content)

    def get(self, key, default=None):
        for item in self.content:
            result = item.get(key, default)
            if result is not None and result is not default:
                return result
        return default

    def __str__(self):
        return '\r\n\r\n'.join(['\r\n'.join([(key + ': ' + value) for key, value in item.items()]) for item in self.content]) 